#include "MotionPlanners/PRM/VisPRM/OOPSMPVisPRM.H"

BeginImplementInstanceFactory(VisPRM, PRMFactory);
EndImplementInstanceFactory(VisPRM);

OOPSMP::VisPRM::VisPRM() : OOPSMP::PRM()
{
}

OOPSMP::VisPRM::~VisPRM( )
{
}

void OOPSMP::VisPRM::clear()
{
    PRM::clear();

    m_undirected  = true;
    m_allowCycles = false;

    max_tries = 1000;
}

void OOPSMP::VisPRM::preprocess(const double rtime)
{ 
    // useful objects for allocating, sampling states
    // and creating local paths
    CoreRobotData_t     crd     = getCoreRobotData();  
    ss      = crd->getStateSpace();                   // State Space
    vss     = crd->getValidStateSampler();            // Valid State Sampler
    lp      = getCoreRobotData()->getLocalPlanner();  // Local Planner

    // initial number of efforts to add a guard/roadmap node is 0
    int nr_tries = 0;

    g_vis = new GuardVisibility( );

    // keep trying while we have not exceeded the 
    // maximum number of failed attempts to add a guard
    while( nr_tries < max_tries )
    {
	// sample a random valid state
	sampleState( );

	// compute the nodes from different connected components that can view the sample
	computeGuardVisibility( );

	if( g_vis->isGuard() || g_vis->isConnector() ) // this is a guard or a "connection" node
	{
	    nr_tries = 0;
	    addGuard();
	}
	else
	{
	    ss->freeState( state );
	    nr_tries++;                 // another failed attempt to add a new node
	}

	g_vis->clearGuardVisibility();
    }
    
    delete( g_vis );
}

// sample a random valid state
void OOPSMP::VisPRM::sampleState( )
{
    state   = ss->allocState();   
    vss->sample( state );

//    printf("SAMPLED STATE: %p\n",state);
}

// is the sampled state already visible from a roadmap on the same
// component as the the node: "node_index"?
bool OOPSMP::VisPRM::isNodeVisibleFromComponent( int node_index )
{
    // is the state already visible by a node in the same component as the i-th node?
    bool not_visible_from_component = true;
    
    for( int i=0; not_visible_from_component == true && i<g_vis->nr_connections; i++ )
	if( m_disjointSet->same( g_vis->connections[i], node_index ) == true )
	    not_visible_from_component = false;

    return not_visible_from_component;
}

// computes the guards from different connected components that can view the state
void OOPSMP::VisPRM::computeGuardVisibility( )
{   
    // current number of guards / roadmap nodes
    int nrVertices  = m_roadmap->getNrVertices();
    
    // a temporary variable to remember paths
    Path_t path = NULL;

    // iterate over all nodes
    for( int i=0; i<nrVertices; i++ )
    {	  
	State_t vTo = (State_t) ((VertexData_t) m_roadmap->getVertexData(i))->m_x;
	
	// is the state already visible by a node in the same component as the i-th node?
	bool notVisibleFromComponent = isNodeVisibleFromComponent( i );  
	
	// consider the i-th roadmap node only if the state is not already 
	// visible by a node in the same component like the i-th node
	if( notVisibleFromComponent == true )
	{
	    //  check whether the i-th node is able to "view" the state
	    if( (path = lp->generateValidForwardConnectPath( state, vTo ) ) )
		g_vis->addConnection( i, path );
	}
    }
}

// add the sampled state as a guard
void OOPSMP::VisPRM::addGuard()
{
    int node_index = m_roadmap->getNrVertices();
    addSample( state );	
    for( int i=0; i<g_vis->nr_connections; i++ )
	addEdge( node_index, g_vis->connections[i], g_vis->paths[i] );  
}

